package ru.ulstu.extractor.service;

import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.ListBranchCommand;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.diff.DiffFormatter;
import org.eclipse.jgit.internal.storage.file.FileRepository;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.revwalk.RevCommit;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.model.Branch;
import ru.ulstu.extractor.model.Changes;
import ru.ulstu.extractor.model.Commit;
import ru.ulstu.extractor.model.FileChange;
import ru.ulstu.extractor.model.LineChange;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static org.apache.logging.log4j.util.Strings.isBlank;

@Service
public class GitRepositoryService {
    @Value("${extractor.custom-projects-dir}")
    private String customProjectsDir;

    public void cloneOrUpdateRepo(String url) throws GitAPIException, IOException {
        Git git;
        if (projectDirExists(getProjectDirectoryFile(url))) {
            Repository localRepo = new FileRepository(getProjectGitDirectory(url));
            git = new Git(localRepo);
            git.pull().call();
        } else {
            git = Git.cloneRepository()
                    .setURI(url)
                    .setDirectory(getProjectDirectoryFile(url))
                    .call();
        }
        Iterable<RevCommit> commits = git.log().call();
        //commits.forEach(c -> System.out.println(c.getFullMessage()));
    }

    private String getProjectDirectory(String url) {
        return (isBlank(customProjectsDir)
                ? System.getProperty("java.io.tmpdir")
                : customProjectsDir) + url.substring(url.lastIndexOf('/'));
    }

    private String getProjectGitDirectory(String url) {
        return getProjectDirectory(url) + "/.git";
    }

    private File getProjectDirectoryFile(String url) {
        return Path.of(getProjectDirectory(url))
                .toFile();
    }

    private boolean projectDirExists(File file) {
        return file.exists();
    }

    public List<Commit> getCommits(String url) throws GitAPIException, IOException {
        cloneOrUpdateRepo(url);
        Repository localRepo = new FileRepository(getProjectGitDirectory(url));
        Git git = new Git(localRepo);
        git.pull().call();
        List<RevCommit> commits = new ArrayList<>();
        //TODO: сделать преобразование в коллекцию "наших объектов"
        git.log().call().forEach(commits::add);

        List<Commit> list = new ArrayList<>();
        RevCommit prevCommit = null;
        for (RevCommit revCommit : commits) {
            Commit commit = new Commit(
                    revCommit.getFullMessage(),
                    revCommit.getAuthorIdent().getName(),
                    Date.from(Instant.ofEpochSecond(revCommit.getCommitTime())));
            if (prevCommit != null) {
                commit.setChanges(findDiffBetweenTwoRevisions(revCommit, prevCommit, localRepo));
            }
            list.add(commit);
            prevCommit = revCommit;
        }
        return list;
    }

    public Changes findDiffBetweenTwoRevisions(RevCommit laterCommit, RevCommit earlierCommit, Repository localRepo) {
        if (laterCommit == null || earlierCommit == null) {
            return null;
        }
        String output = null;
        try (ByteArrayOutputStream out = new ByteArrayOutputStream()) {
            DiffFormatter diffFormatter = new DiffFormatter(out);
            diffFormatter.setRepository(localRepo);
            diffFormatter.format(earlierCommit.getId(), laterCommit.getId());
            output = out.toString();
        } catch (IOException e) {
            throw new RuntimeException("Error occurred during diff computation. Message: " + e.getMessage());
        }
        return parseOutputDiff(output);
    }

    private Changes parseOutputDiff(String output) {
        Changes changes = new Changes();
        String[] strings = output.split("\n");
        FileChange fileChange = new FileChange();
        int stringsLength = strings.length - 1;
        for (int i = 0; i < strings.length; i++) {
            while (i < stringsLength) {
                Optional<String> maybeFileName = getFileName(strings[i]);
                if (maybeFileName.isPresent()) {
                    fileChange = new FileChange();
                    fileChange.setFile(maybeFileName.get());
                    /// вытащить другие изменения из коммита
                    changes.getFileChanges().add(fileChange);
                }
                LineChange lineChange = new LineChange();
                if (strings[i].startsWith("-")) {
                    while ((i < stringsLength) && strings[i].startsWith("-")) {
                        if (strings[i].startsWith("---")) {
                            i++;
                        } else {
                            lineChange.setLineFrom(lineChange.getLineFrom() + "\n" + strings[i]);
                            i++;
                        }
                    }
                }
                if (strings[i].startsWith("+")) {
                    while ((i < stringsLength) && strings[i].startsWith("+")) {
                        if (strings[i].startsWith("+++")) {
                            i++;
                        } else {
                            lineChange.setLineTo(lineChange.getLineTo() + "\n" + strings[i]);
                            i++;
                        }
                    }
                }
                if ((lineChange.getLineTo() != null) || (lineChange.getLineFrom() != null)
                        || ((lineChange.getLineTo() != null) && (lineChange.getLineFrom() != null))) {
                    fileChange.getLineChanges().add(lineChange);
                } else {
                    i++;
                }
            }
        }
        return changes;
    }

    private Optional<String> getFileName(String commitString) {
        String startString = "diff --git a/";
        if (commitString.startsWith(startString)) {
            String name = commitString.substring(commitString.indexOf(startString) + startString.length(), commitString.indexOf(" b/"));
            return Optional.of(name);
        }
        return Optional.empty();
    }

    private Optional<String> getLineTo(String commitString) {
        String startString = "+";
        String startPlus = "+++";
        if (commitString.startsWith(startString)) {
            String lineTo = commitString.substring(commitString.indexOf(startString) + startString.length());
            if (commitString.startsWith(startPlus)) {
                return Optional.empty();
            } else {
                return Optional.of(lineTo);
            }
        }
        return Optional.empty();
    }

    private Optional<String> getLineFrom(String commitString) {
        String startString = "-";
        String startMinus = "---";
        if (commitString.startsWith(startString)) {
            String lineFrom = commitString.substring(commitString.indexOf(startString) + startString.length());
            if (commitString.startsWith(startMinus)) {
                return Optional.empty();
            } else {
                return Optional.of(lineFrom);
            }
        }
        return Optional.empty();
    }

    public List<Branch> getBranches(String url) throws GitAPIException, IOException {
        cloneOrUpdateRepo(url);
        Repository localRepo = new FileRepository(getProjectGitDirectory(url));
        Git git = new Git(localRepo);
        return git.branchList().setListMode(ListBranchCommand.ListMode.REMOTE)
                .call()
                .stream()
                .map(r -> new Branch(r.getName()))
                .collect(Collectors.toList());
    }
}
